#ifndef ANIMATION_HPP
#define ANIMATION_HPP

#include <vector>
#include <GLFW/glfw3.h>
#include "Interpolations.hpp"
template <class myType>
class Animation
{
public:
	Animation(){ init(0.0,0.0);}
	Animation(double EndTime){init(0.0,EndTime);}
	Animation(double StartTime, double EndTime){init(StartTime,EndTime);}
	Animation(double StartTime, double EndTime, Renderable * r){init(StartTime,EndTime); obj = r;};
	int addValuePair(myType start, myType end){return addValuePair(start,end,start,end);}
	int addValuePair(myType start, myType end,myType start_, myType end_)
	{
		startValues.push_back(start);
		endValues.push_back(end);
		startValues_D.push_back(start_);
		endValues_D.push_back(end_);
		return startValues.size()-1;
	}
	bool hasStarted() {return hasStarted(glfwGetTime());}
	bool hasStarted(double time) { return time > startTime;}
	bool isDone(){ return isDone(glfwGetTime());}
	bool isDone(double time) { return (time > endTime);}
	bool isRunning() { return isRunning(glfwGetTime());}
	bool isRunning(double time) {return hasStarted(time) && !isDone(time);}
	std::vector<myType> getValues() { return getValues(glfwGetTime()); }
	std::vector<myType> getValues(double time) { return getValues(time,defaultInterpolation);	}
	std::vector<myType> getValues(double time, myType (*Interpolation)(myType start, myType end,myType start_, myType end_, float progress))
	{
		if(time < startTime)
			return startValues;
		if(time > endTime)
			return endValues;
		double progress = (time - startTime)/(endTime - startTime);
		std::vector<myType> ret;
		for(unsigned int i = 0; i < startValues.size(); i++)
		{
			ret.push_back(Interpolation(startValues[i],endValues[i],startValues_D[i],endValues_D[i],(float)progress));
		}
		return ret;
	}

	void setDefaultInterpolation(myType (*Interpolation)(myType start, myType end,myType start_, myType end_, float progress))
	{
		defaultInterpolation = Interpolation;
	}

	void useCubicInterpolation() { setDefaultInterpolation(&(cubicInterpolation<myType>));}
	void useLinearInterpolation() { setDefaultInterpolation(&(linearInterpolation<myType>));}

	double getStartTime() { return startTime };

	void setObject(Renderable * renderable) { obj = renderable;}
	Renderable* getObject() { return obj; }

private:
	double startTime;
	double endTime;
	Renderable * obj;

	void init(double Start,double End)
	{
		//typedef myType T;
		startTime=Start;
		endTime=End;
		setDefaultInterpolation(&(linearInterpolation<myType>));
	}

	
	
	myType (*defaultInterpolation)(myType start, myType end, myType start_, myType end_, float progress);
	
	std::vector<myType> startValues;
	//1st derivatives at start points
	std::vector<myType> startValues_D;
	std::vector<myType> endValues;
	//1st derivatives at end points
	std::vector<myType> endValues_D;

};

#endif