#ifndef CAMERA_H
#define CAMERA_H

#include <glm/glm.hpp>

class Camera
{
public:
	Camera();
	~Camera();

	const glm::vec3& position() const;
	glm::mat4 orientation() const;

	glm::vec3 forward() const;
	glm::vec3 right() const;
	glm::vec3 up() const;

	glm::mat4 matrix() const;
	glm::mat4 projection() const;
	glm::mat4 view() const;

	float fieldOfView() const;
	float nearPlane() const;
	float farPlane() const;
	float aspectRatio() const;


	void setPosition(const glm::vec3& position);
	void offsetPosition(const glm::vec3& offset);

	void setFieldOfView(float fieldOfView);
	void setPlanes(float nearPlane, float farPlane);

	void setOrientation(glm::mat4 o);
	void setOrientation(float up, float right);
	void offsetOrientation(float up, float right);
	
	void lookAt(glm::vec3 position);

	void setAspectRatio(float aspectRatio);
	void printAngles();

private:
	glm::vec3 mPosition;
	glm::mat4 mOrientation;
	float mHorizontalAngle;
	float mVerticalAngle;
	float mFieldOfView;
	float mNearPlane;
	float mFarPlane;
	float mAspectRatio;

	void normalizeAngles();
};

#endif