#include <vector>
#include <GL/glew.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
using namespace glm;
#include <assimp/types.h>
#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>
#include <string>
#include "main.h"
#include "Renderable.hpp"
#include "shader.hpp"

Renderable::~Renderable()
{
    //TODO Cleanup
	//ACHTUNG!
	//NICHT hier deleteBuffers aufrufen!
	//sonst werden die Buffer nach dem ersten kopieren des Objekts gelscht!

    vbo_colors.clear();
    vbo_textureCoordinates.clear();
    vbo_vertices.clear();
	idx_vertices.clear();
	vbo_normals.clear();

    textureId=0;
	useModelTextures = false;
    tessellation=false;
	visible=false;
	reflection = false;
	
	tessellation_Parameters.vals[0] = 0.0f;
	tessellation_Parameters.vals[1] = 0.0f;
	tessellation_Parameters.vals[2] = 0.0f;
	tessellation_Parameters.vals[3] = 0.0f;
	tessellation_Parameters.vals[4] = 0.0f;

	vertexbuffer=0;
	idx_vertexbuffer=0;
	uvbuffer=0;
	colorBuffer=0;
	normalsBuffer=0;
	
}

void Renderable::deleteBuffers()
{
	glDeleteBuffers(1,&vertexbuffer);
	glDeleteBuffers(1,&idx_vertexbuffer);
	glDeleteBuffers(1,&uvbuffer);
	glDeleteBuffers(1,&colorBuffer);
	glDeleteBuffers(1,&normalsBuffer);
}
Renderable::Renderable(const Renderable& a)
{
    tessellation=a.tessellation;
	tessellation_Parameters.vals[0] = a.tessellation_Parameters.vals[0];
	tessellation_Parameters.vals[1] = a.tessellation_Parameters.vals[1];
	tessellation_Parameters.vals[2] = a.tessellation_Parameters.vals[2];
	tessellation_Parameters.vals[3] = a.tessellation_Parameters.vals[3];
	tessellation_Parameters.vals[4] = a.tessellation_Parameters.vals[4];
    name=a.name;
    scaleMatrix = a.scaleMatrix;
	rotationMatrix = a.rotationMatrix;
	translationMatrix = a.translationMatrix;
    position=a.position;
	angles=a.angles;
	visible=a.visible;
	depth_test=a.depth_test;
	reflection = a.reflection;
    vbo_vertices.insert(vbo_vertices.end(),a.vbo_vertices.begin(),a.vbo_vertices.end());
	idx_vertices.insert(idx_vertices.end(),a.idx_vertices.begin(),a.idx_vertices.end());
    vbo_colors.insert(vbo_colors.end(),a.vbo_colors.begin(),a.vbo_colors.end());
    vbo_textureCoordinates.insert(vbo_textureCoordinates.end(),a.vbo_textureCoordinates.begin(),a.vbo_textureCoordinates.end());
    vbo_normals.insert(vbo_normals.end(),a.vbo_normals.begin(),a.vbo_normals.end());

    textureId=a.textureId;
    vertexbuffer=a.vertexbuffer;
	idx_vertexbuffer=a.idx_vertexbuffer;
    uvbuffer=a.uvbuffer;
    colorBuffer=a.colorBuffer;
    normalsBuffer=a.normalsBuffer;

	
}
Renderable::Renderable()
{
    name=("");
    scaleMatrix = glm::mat4(1.0f);
	rotationMatrix = glm::mat4(1.0f);
	translationMatrix = glm::mat4(1.0f);
    position = glm::vec3(0);
	angles = glm::vec3(0);
	reflection = false;
	tessellation_Parameters.vals[0] = 1.0f;
	tessellation_Parameters.vals[1] = 1.0f;
	tessellation_Parameters.vals[2] = 1.0f;
	tessellation_Parameters.vals[3] = 1.0f;
	tessellation_Parameters.vals[4] = 1.0f;
}
Renderable::Renderable(std::vector<glm::vec3> vertices, std::vector<glm::vec2> texCoords,GLuint texId)
: textureId(texId)
{
    name=("");
	tessellation=false;
	useModelTextures=true;
	visible=true;
	reflection = false;
	depth_test=true;
    scaleMatrix = glm::mat4(1.0f);
	rotationMatrix = glm::mat4(1.0f);
	translationMatrix = glm::mat4(1.0f);
    if(vertices.size() != texCoords.size())
        return;
    vbo_vertices.insert(vbo_vertices.end(),vertices.begin(),vertices.end());
    vbo_textureCoordinates.insert(vbo_textureCoordinates.end(),texCoords.begin(),texCoords.end());
    for(unsigned int i=0;i<vertices.size();i++)
	{
        vbo_colors.push_back(glm::vec3(1));
		idx_vertices.push_back(i);
		if(i%3==2)
		{
			//calc normalvectors:
			vec3 v1=vbo_vertices[i-2];
			vec3 v2=vbo_vertices[i-1];
			vec3 v3=vbo_vertices[i];
			vec3 norm = glm::cross((v2-v1),(v3-v1));
			//add 3 times
			vbo_normals.push_back(norm);
			vbo_normals.push_back(norm);
			vbo_normals.push_back(norm);
		}
	}
}
Renderable::Renderable(std::vector<glm::vec3> vertices, std::vector<glm::vec3> colorValues)
{
    name=("");
    tessellation=false;
	useModelTextures=false;
	visible=true;
	depth_test=true;
	reflection = false;
    textureId=0;
    scaleMatrix = glm::mat4(1.0f);
	rotationMatrix = glm::mat4(1.0f);
	translationMatrix = glm::mat4(1.0f);
    position = glm::vec3(0);
	angles = glm::vec3(0);
    if(vertices.size() != colorValues.size())
        return;
    vbo_vertices.insert(vbo_vertices.end(),vertices.begin(),vertices.end());
    vbo_colors.insert(vbo_colors.end(),colorValues.begin(),colorValues.end());
    for(unsigned int i=0;i<vertices.size();i++)
	{
		idx_vertices.push_back(i);
        vbo_textureCoordinates.insert(vbo_textureCoordinates.end(),glm::vec2(1));
		if(i%3==2)
		{
			//calc normalvectors:
			vec3 v1=vbo_vertices[i-2];
			vec3 v2=vbo_vertices[i-1];
			vec3 v3=vbo_vertices[i];
			vec3 norm = glm::cross((v2-v1),(v3-v1));
			//add 3 times
			vbo_normals.push_back(norm);
			vbo_normals.push_back(norm);
			vbo_normals.push_back(norm);
		}
	}

}
Renderable::Renderable(const std::string path)
{
    tessellation=false;
	useModelTextures=false;
	visible=true;
	depth_test=true;
	reflection = false;
    name=("");
	scaleMatrix = glm::mat4(1.0f);
	rotationMatrix = glm::mat4(1.0f);
	translationMatrix = glm::mat4(1.0f);
    textureId=0;
    Assimp::Importer importer;
    const aiScene* scene;
    printf(path.c_str());
    scene = importer.ReadFile( path, aiProcessPreset_TargetRealtime_Quality);
    std::vector<glm::vec3> vert;
	std::vector<glm::vec3> uv_channel;
    std::vector<glm::vec3> norm;
    if(scene->HasMeshes())
    {
        //printf("%d\n",scene->mNumMeshes);
        for(unsigned int i=0;i<scene->mNumMeshes;i++)
        {
            const struct aiMesh* mesh = scene->mMeshes[i];
            if(mesh->HasPositions())
            {
               // printf("%d\n",mesh->mNumVertices);
                for(unsigned int j=0;j<mesh->mNumVertices;j++)
                {
                    aiVector3D v=mesh->mVertices[j];
					aiVector3D u=mesh->mTextureCoords[0][j];
                    glm::vec3 nv=glm::vec3(v.x,v.y,v.z);
					vbo_vertices.push_back(nv);
					vbo_textureCoordinates.push_back(glm::vec2(u.x,u.y));
					vbo_colors.push_back(glm::vec3(1));
                    if(mesh->HasNormals())
                    {
                        aiVector3D no=mesh->mNormals[j];
                        glm::vec3 nno=glm::vec3(no.x,no.y,no.z);
						vbo_normals.push_back(nno);
                    }
                }
                for(unsigned int j=0;j<mesh->mNumFaces;j++)
                {
                    const aiFace f=mesh->mFaces[j];
                    unsigned int i0=f.mIndices[0];
                    unsigned int i1=f.mIndices[1];
                    unsigned int i2=f.mIndices[2];

					idx_vertices.push_back(i0);
					idx_vertices.push_back(i1);
					idx_vertices.push_back(i2);

					//TODO Calculate normals wenn nicht gegeben
					
                }
            }

        }
    }//Ende Scene->HasMeshes()
	//TODO Read Textures from 3ds file
	if(scene->HasTextures())
	{
		for(unsigned int i=0;i<scene->mNumTextures;i++)
		{
			aiTexture* t = scene->mTextures[i];
			
		}
	}
}
void Renderable::setTessellationParameterFixed(float inner_l,float inner_u, float outer0, float outer1, float outer2)
{
	tessellation_Parameters.floats.InnerLevel_lower=inner_l;
	tessellation_Parameters.floats.InnerLevel_upper=inner_u;
	tessellation_Parameters.floats.OuterLevel0=outer0;
	tessellation_Parameters.floats.OuterLevel1=outer1;
	tessellation_Parameters.floats.OuterLevel2=outer2;
}
void Renderable::setDepthTest(bool f)
{
	depth_test=f;
}
bool Renderable::getDepthTest()
{
	return depth_test;
}
bool Renderable::getReflection()
{
	return reflection;
}
void Renderable::setReflection(bool f)
{
	reflection=f;
}
void Renderable::setVisible(bool v)
{
	visible=v;
}
void Renderable::setModelTexturesFlag(bool f)
{
	useModelTextures=true;
}
bool Renderable::isTessellated()
{
    return tessellation;
}
void Renderable::setTessellation(bool tes)
{

    tessellation=tes;
}
void Renderable::bufferData(){
    glGenBuffers(1, &vertexbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
	glBufferData(GL_ARRAY_BUFFER, vbo_vertices.size() * sizeof(glm::vec3), &vbo_vertices[0], GL_STATIC_DRAW);

	glGenBuffers(1, &uvbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, uvbuffer);
	glBufferData(GL_ARRAY_BUFFER, vbo_textureCoordinates.size() * sizeof(glm::vec2), &vbo_textureCoordinates[0], GL_STATIC_DRAW);

	glGenBuffers(1,&colorBuffer);
	glBindBuffer(GL_ARRAY_BUFFER,colorBuffer);
	glBufferData(GL_ARRAY_BUFFER,vbo_colors.size()*sizeof(glm::vec3),&vbo_colors[0],GL_STATIC_DRAW);

	glGenBuffers(1,&normalsBuffer);
	glBindBuffer(GL_ARRAY_BUFFER,normalsBuffer);
	glBufferData(GL_ARRAY_BUFFER,vbo_normals.size()*sizeof(glm::vec3),&vbo_normals[0],GL_STATIC_DRAW);

	glGenBuffers(1, &idx_vertexbuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, idx_vertexbuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, idx_vertices.size() * sizeof(unsigned int), &idx_vertices[0], GL_STATIC_DRAW);

}
void Renderable::addNormals(std::vector<glm::vec3> norms)
{
    vbo_normals.clear();
    vbo_normals.insert(vbo_normals.end(),norms.begin(),norms.end());
}
GLfloat* Renderable::getTessellationParamsOuterVector()
{
	return &(tessellation_Parameters.vals)[2];
}
GLfloat* Renderable::getTessellationParamsInnerVector()
{
	return &(tessellation_Parameters.vals)[0];
}
void Renderable::addColors(std::vector<glm::vec3> cols)
{
    vbo_colors.clear();
    vbo_colors.insert(vbo_colors.end(),cols.begin(),cols.end());
}
glm::mat4 Renderable::getModelMatrix()
{
	return scaleMatrix*translationMatrix*rotationMatrix;
}
void Renderable::setRotation(glm::vec3 ang)
{
	glm::vec3 rot = ang-angles;
	rotationMatrix = glm::rotate(glm::mat4(1),rot.x,glm::vec3(1,0,0));
	rotationMatrix = glm::rotate(rotationMatrix,rot.y,glm::vec3(0,1,0));
	rotationMatrix = glm::rotate(rotationMatrix,rot.z,glm::vec3(0,0,1));
	angles = ang;
}
void Renderable::setPosition(glm::vec3 pos)
{
    glm::vec3 trans = pos-position;
	translationMatrix = glm::translate(glm::mat4(1), pos); // Create our view matrix which will translate us to the new position
	//ModelMatrix = glm::translate(ModelMatrix, pos);
    position = pos;
}
void Renderable::setScale(glm::vec3 scale)
{
	//ModelMatrix = glm::scale(ModelMatrix,scale);
	
	scaleMatrix = glm::scale_slow(glm::mat4(1),scale);
}
glm::vec3 Renderable::getPosition()
{
	return position;
}
void Renderable::setName(std::string n)
{
    name=n;
}
std::string Renderable::getName()
{
    return name;
}
bool Renderable::isTextured()
{
    return textureId!=0;
}
void Renderable::setTextureId(GLuint texId)
{
	textureId = texId;
}
GLuint Renderable::getTextureId()
{
    return textureId;
}
void Renderable::bindBuffers() {this->bindBuffers(0);}
void Renderable::bindBuffers(int pass)
{
	glEnableVertexAttribArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
    glVertexAttribPointer(
        0,                  // attribute 0. No particular reason for 0, but must match the layout in the shader.
        3,                  // size
        GL_FLOAT,           // type
        GL_FALSE,           // normalized?
        0,                  // stride
        (void*)0            // array buffer offset
    );
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, idx_vertexbuffer);
	if(pass != 0)
	{
		glEnableVertexAttribArray(1);
		glBindBuffer(GL_ARRAY_BUFFER,colorBuffer);
		glVertexAttribPointer(
			1,
			3,
			GL_FLOAT,
			GL_FALSE,
			0,
			(void*)0
		);

		glEnableVertexAttribArray(2);
		glBindBuffer(GL_ARRAY_BUFFER, uvbuffer);
		glVertexAttribPointer(
			2,                                // attribute. No particular reason for 1, but must match the layout in the shader.
			2,                                // size : U+V => 2
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
		);

		glEnableVertexAttribArray(3);
		glBindBuffer(GL_ARRAY_BUFFER, normalsBuffer);
		glVertexAttribPointer(
			3,                                // attribute. No particular reason for 1, but must match the layout in the shader.
			3,                                // size : vector normal
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
		);
	}

}
void Renderable::bindTexture()
{
	glBindTexture(GL_TEXTURE_2D, textureId);
}
void Renderable::setColor(glm::vec3 col)
{
    vbo_colors.clear();
    std::vector<glm::vec3>::iterator it;
    for(unsigned int i=0;i<vbo_vertices.size();i++)
    {
        vbo_colors.push_back(col);
    }
    //std::vector<glm::vec3>::iterator it=Renderable::vbo_colors.at(vertexPosition);
}
void Renderable::render(){ this->render(0);}
void Renderable::render(int renderpass)
{
    //printf("Rendering from %d to %d",vertexPosition,vertexPosition+vertexCounter);

	if(visible)
	{
		//if(!tessellation)
		if(!TESSELLATION || renderpass==0)
		{
			//glDrawArrays(GL_TRIANGLES, 0, vbo_vertices.size());
			glDrawElements(GL_TRIANGLES, idx_vertices.size(), GL_UNSIGNED_INT, 0);
		}
		else
		{
			glPatchParameteri( GL_PATCH_VERTICES, 3);
			//glDrawArrays(GL_PATCHES, 0,vbo_vertices.size());
			glDrawElements(GL_PATCHES, idx_vertices.size(), GL_UNSIGNED_INT, 0);/**/
		}
	}
}