#ifndef RENDERABLE_HPP
#define RENDERABLE_HPP
#include <string>

class Renderable
{


  public:                              // ffentlich
    Renderable();                      // der Default-Konstruktor
    Renderable(std::vector<glm::vec3> vertices, std::vector<glm::vec2> texCoords,GLuint texId);               // weiterer Konstruktor mit Parameter und Defaultwert
    Renderable(std::vector<glm::vec3> vertices, std::vector<glm::vec3> colorValues);
    Renderable(const std::string path);               // weiterer Konstruktor mit Parameter und Defaultwert

    Renderable(const Renderable& a);   // Copy-Konstruktor
    ~Renderable();                     // der Destruktor
	


    void setColor(glm::vec3 col);
    glm::mat4 getModelMatrix();
    void setPosition(glm::vec3 pos);
	glm::vec3 getPosition();
	void setRotation(glm::vec3 ang);
	void setScale(glm::vec3 scale);
	void setVisible(bool v);
	void setTessellationParameterFixed(float inner_l, float inner_u ,float outer0,float outer1,float outer2);
	
    bool isTextured();                  //ist das Object texturiert
    bool isTessellated();
    void setTessellation(bool tes);
	void setTextureId(GLuint texId);
    GLuint getTextureId();
	void render();
    void render(int pass);
    std::string getName();
    void setName(std::string);
    void bindBuffers();
	void bindBuffers(int pass);
	void bindTexture();

	void deleteBuffers();

	void setModelTexturesFlag(bool f);
	void setDepthTest(bool f);
	bool getDepthTest();
	void setReflection(bool f);
	bool getReflection();

    void addNormals(std::vector<glm::vec3> norms);
    void addColors(std::vector<glm::vec3> cols);
    void bufferData();

	GLfloat* getTessellationParamsInnerVector();
	GLfloat* getTessellationParamsOuterVector();
    std::vector<glm::vec3>    vbo_vertices;
	std::vector<unsigned int> idx_vertices;
    std::vector<glm::vec2>    vbo_textureCoordinates;
    std::vector<glm::vec3>    vbo_colors;
    std::vector<glm::vec3>    vbo_normals;


  private:                             // privat

  //  int vertexCounter;
    bool tessellation;
	bool depth_test;
	bool useModelTextures;
	bool visible;
	bool reflection;
	union Parameters {
		struct fl{
			float InnerLevel_lower,InnerLevel_upper,OuterLevel0,OuterLevel1,OuterLevel2;
		} floats;
		float vals[5];
	} tessellation_Parameters;

    std::string name;
    GLuint textureId;
    glm::mat4 translationMatrix;
	glm::mat4 rotationMatrix;
	glm::mat4 scaleMatrix;
    glm::vec3 position;
	glm::vec3 angles;

    GLuint vertexbuffer;
    GLuint uvbuffer;
    GLuint colorBuffer;
    GLuint normalsBuffer;

	GLuint idx_vertexbuffer;
	GLuint idx_uvbuffer;
	GLuint idx_colorBuffer;
	GLuint idx_normalsBuffer;
};

#endif
