#include <stdlib.h>
#include <vector>
#include <iostream>
#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>
#include <glm/gtx/euler_angles.hpp>
#include <glm/gtx/norm.hpp>
#include <glm/gtx/string_cast.hpp>

using namespace glm;

#include "quaternion_utils.hpp"

#include <png.h>
#include "projectConstants.h"
#include "main.h"
//shader.hpp und shader.cpp kommen von www.opengl-tutorial.org
#include "shader.hpp"
//objloader.hpp und objloader.cpp kommen von www.opengl-tutorial.org
//#include "objloader.hpp"
//texture.hpp und texture.cpp kommen von www.opengl-tutorial.org
#include "texture.hpp"
#include "Renderable.hpp"
#include "Szene.hpp"

#include "Camera.h"
#include "Animation.hpp"

int frame = 0;
int width, height;
GLuint firstPass_1_ProgramID;
GLuint firstPass_2_ProgramID;
GLuint secondPass_ProgramID;
GLuint thirdPass_ProgramID;

std::string project::title =  ( "GiSt");
std::string project::base  =  (".");
std::string project::res   =  base+SEP+("res");
std::string project::models=  res + SEP + ("models");
std::string project::textures = res + SEP + ("textures");
std::string project::src  = base+SEP+("src");
int project::init_breite = 1024;
int project::init_hoehe = 768;
GLFWwindow* project::window = 0;
glm::mat4 project::ProjectionMatrix;
glm::mat4 project::ModelMatrix;
glm::mat4 project::ViewMatrix;

double project::last_frame_time=0;
float project::bewegungs_geschwindigkeit = 10.0f;
glm::vec3 project::position = glm::vec3(0.0f,0.0f,-10.0f);
float project::blickrichtung  =  0.0f;//PI/2.0f; //positive Z-Richtung in Radiant
float project::neigungswinkel  = 0.0f;
//glm::vec3 project::light_source_position=glm::vec3(0.0f,5.0f,0.0f);

Camera camera;
const float moveSpeed = 20.0f;

Camera lightCamera_pos_x;
Camera lightCamera_neg_x;
Camera lightCamera_pos_y;
Camera lightCamera_neg_y;
Camera lightCamera_pos_z;
Camera lightCamera_neg_z;

const int shadowTexSideLength = 1024/2;

const int rPass2Width = 1024;
const int rPass2Height = 1024;

const float cm_cams_near_plane = 0.1;
const float cm_cams_far_plane = 300.0;

const double fps_update_interval_in_sec = 0.3;
double fps_last_update = 0.0;
int fps_framecounter = 0;
bool fps_showFPS = true;
bool fps_key_pressed = false;

bool wire_key_pressed = false;
bool wire_showWireFrame = false;

bool varShadow_key_pressed = false;
int varShadow_mode = 1;

bool doThirdPass= true;

/* 3rdRenderpass Vertices */
GLuint vbo_fbo_vertices;

std::vector<Animation<vec3>> animationVec;
std::vector<Animation<vec3>> planeAnimationVec;
double wireTogTime;
double shadowTogStartTime;
double shadowTogEndTime;
bool doAnimations = true;
bool animation_key_pressed = false;

std::vector<float> interpol;

std::vector<Animation<vec3>> cameraMovementVec;

unsigned char* loadImg(const char* imagepath)
{
	// Data read from the header of the BMP file
	unsigned char header[54];
	unsigned int dataPos;
	unsigned int imageSize;
	unsigned int width, height;
	// Actual RGB data
	unsigned char * data;

	// Open the file
	FILE * file = fopen(imagepath,"rb");
	if (!file)							    {printf("Image could not be opened\n"); return 0;}

	// Read the header, i.e. the 54 first bytes

	// If less than 54 byes are read, problem
	if ( fread(header, 1, 54, file)!=54 ){ 
		printf("Not a correct BMP file1\n");
		return false;
	}
	// A BMP files always begins with "BM"
	if ( header[0]!='B' || header[1]!='M' ){
		printf("Not a correct BMP file2\n");
		return 0;
	}
	// Make sure this is a 24bpp file
	if ( *(int*)&(header[0x1E])!=0  )         {printf("Not a correct BMP file3\n");    return 0;}
	if ( *(int*)&(header[0x1C])!=24 )         {printf("Not a correct BMP file4\n");    return 0;}

	// Read the information about the image
	dataPos    = *(int*)&(header[0x0A]);
	imageSize  = *(int*)&(header[0x22]);
	width      = *(int*)&(header[0x12]);
	height     = *(int*)&(header[0x16]);

	// Some BMP files are misformatted, guess missing information
	if (imageSize==0)    imageSize=width*height*3; // 3 : one byte for each Red, Green and Blue component
	if (dataPos==0)      dataPos=54; // The BMP header is done that way

	// Create a buffer
	data = new unsigned char [imageSize];
	fread(data,1,imageSize,file);

	// Everything is in memory now, the file wan be closed
	fclose (file);
	return data;
}

int main()
{
    project::init();

    //int     frame = 0;
    bool    running = true;
 	//firstPass_ProgramID = LoadShaders( (project::src+SEP+"SimpleVertexShader.vertexshader").data(), (project::src+SEP+("SimpleFragmentShader.fragmentshader")).data() );
 	firstPass_1_ProgramID = LoadShaders( (project::src+SEP+"1stPass.vertexshader").data(), 0,
							0,(project::src+SEP+("1stPass.geometryshader")).data(),(project::src+SEP+("1stPass_1.fragmentshader")).data() );
	bool envMappingExtreme = false;
	if(envMappingExtreme)
	{
		firstPass_2_ProgramID = LoadShaders( (project::src+SEP+"1stPass.vertexshader").data(), (project::src+SEP+("1stPass_2.tesControlShader")).data(),
							(project::src+SEP+("1stPass_2.tesEvaluationShader")).data(),(project::src+SEP+("1stPass.geometryshader")).data(),(project::src+SEP+("1stPass_2.fragmentshader")).data() );
	}
	else
	{
		firstPass_2_ProgramID = LoadShaders( (project::src+SEP+"1stPass.vertexshader").data(), 0,
							0,(project::src+SEP+("1stPass.geometryshader")).data(),(project::src+SEP+("1stPass_2.fragmentshader")).data() );
	}


	secondPass_ProgramID = LoadShaders( (project::src+SEP+"2ndPass.vertexshader").data(),(project::src+SEP+("2ndPass.tesControlShader")).data(),
							(project::src+SEP+("2ndPass.tesEvaluationShader")).data(),0, (project::src+SEP+("2ndPass.fragmentshader")).data() );

	thirdPass_ProgramID = LoadShaders( (project::src+SEP+"3rdPass.vertexshader").data(),0,
                           0,0,(project::src+SEP+("3rdPass.fragmentshader")).data() );

    Szene::init();
	/*camera.offsetOrientation(0,-90.0f);
	camera.setPosition(glm::vec3(50, 0, 0));*/
	camera.setAspectRatio((float)project::init_breite/(float)project::init_hoehe);


    //GLenum error = glGetError();

    
    GLuint light_source_position_11p  = glGetUniformLocation(firstPass_1_ProgramID, "light_source_position");
	GLuint near_far_plane_11p = glGetUniformLocation(firstPass_1_ProgramID, "near_far_plane");
	GLuint cm_mat_11p = glGetUniformLocation(firstPass_1_ProgramID, "cm_mat");
	GLuint layer_11p = glGetUniformLocation(firstPass_1_ProgramID, "layerToRender");
    GLuint ModelMatrixID_11p = glGetUniformLocation(firstPass_1_ProgramID, "ModelMatrix");
	GLuint CameraPositionID_11p = glGetUniformLocation(firstPass_1_ProgramID, "camera_pos_world");
	GLuint trans_to_ref_11p = glGetUniformLocation(firstPass_1_ProgramID,"trans_to_ref");

	GLuint light_source_position_12p  = glGetUniformLocation(firstPass_2_ProgramID, "light_source_position");
	GLuint near_far_plane_12p = glGetUniformLocation(firstPass_2_ProgramID, "near_far_plane");
	GLuint cm_mat_12p = glGetUniformLocation(firstPass_2_ProgramID, "cm_mat");
	GLuint layer_12p = glGetUniformLocation(firstPass_2_ProgramID, "layerToRender");
    GLuint ModelMatrixID_12p = glGetUniformLocation(firstPass_2_ProgramID, "ModelMatrix");
	GLuint MatrixID_12p = glGetUniformLocation(firstPass_2_ProgramID, "MVP");
    GLuint ViewMatrixID_12p = glGetUniformLocation(firstPass_2_ProgramID, "ViewMatrix");
	GLuint ProjectionMatrixID_12p = glGetUniformLocation(firstPass_2_ProgramID, "proj_mat");
	GLuint CameraPositionID_12p = glGetUniformLocation(firstPass_2_ProgramID, "camera_pos_world");
	GLuint trans_to_ref_12p = glGetUniformLocation(firstPass_2_ProgramID,"trans_to_ref");
	GLuint useTexture_12p = glGetUniformLocation(firstPass_2_ProgramID, "useTexture");
	GLuint tessellation_12p = glGetUniformLocation(firstPass_2_ProgramID, "tessellation");
	GLuint tessellationParam_outer_12p = glGetUniformLocation(firstPass_2_ProgramID, "tessellationControlParameters_outer");
	GLuint tessellationParam_inner_12p = glGetUniformLocation(firstPass_2_ProgramID, "tessellationControlParameters_inner");
	GLuint elapsedTimeID_12p = glGetUniformLocation(firstPass_2_ProgramID, "elapsedTime");
	

	GLuint texSampler_2p  = glGetUniformLocation(secondPass_ProgramID, "myTextureSampler");
	GLuint reflection_flag_2p  = glGetUniformLocation(secondPass_ProgramID, "reflection");
	GLuint tex_size_2p  = glGetUniformLocation(secondPass_ProgramID, "tex_size");
	GLuint light_source_position_2p  = glGetUniformLocation(secondPass_ProgramID, "light_source_position");
	GLuint cm_z_tex_2p = glGetUniformLocation(secondPass_ProgramID, "cm_z_tex");
	GLuint near_far_plane_2p = glGetUniformLocation(secondPass_ProgramID, "near_far_plane");
	GLuint MatrixID_2p = glGetUniformLocation(secondPass_ProgramID, "MVP");
    GLuint ModelMatrixID_2p = glGetUniformLocation(secondPass_ProgramID, "ModelMatrix");
    GLuint ViewMatrixID_2p = glGetUniformLocation(secondPass_ProgramID, "ViewMatrix");
	GLuint ProjectionMatrixID_2p = glGetUniformLocation(secondPass_ProgramID, "proj_mat");
	GLuint CameraPositionID_2p = glGetUniformLocation(secondPass_ProgramID, "camera_pos_world");
	GLuint elapsedTimeID_2p = glGetUniformLocation(secondPass_ProgramID, "elapsedTime");
	GLuint useTexture_2p = glGetUniformLocation(secondPass_ProgramID, "useTexture");
    GLuint tessellation_2p = glGetUniformLocation(secondPass_ProgramID, "tessellation");
	GLuint tessellationParam_outer_2p = glGetUniformLocation(secondPass_ProgramID, "tessellationControlParameters_outer");
	GLuint tessellationParam_inner_2p = glGetUniformLocation(secondPass_ProgramID, "tessellationControlParameters_inner");
	GLuint kernelsize_2p = glGetUniformLocation(secondPass_ProgramID, "kernelsize");
	GLuint mipmapLevel_2p = glGetUniformLocation(secondPass_ProgramID,"mipmap_level");
	GLuint varShadow_2p = glGetUniformLocation(secondPass_ProgramID,"varShadow");
	//Subroutine
	//GLuint subOmni_2p = glGetSubroutineIndex(secondPass_ProgramID, GL_FRAGMENT_SHADER, "omnidirectionalShadowmap");
	//GLuint subvar_2p = glGetSubroutineIndex(secondPass_ProgramID, GL_FRAGMENT_SHADER, "varianceShadowmap");
	//GLuint subCall_2p = glGetSubroutineUniformLocation(secondPass_ProgramID,GL_FRAGMENT_SHADER,"whichShading");

	GLuint VP_3p = glGetUniformLocation(thirdPass_ProgramID, "ViewProj");
	GLuint invVP_3p = glGetUniformLocation(thirdPass_ProgramID, "invViewProj");
	GLuint oldVP_3p = glGetUniformLocation(thirdPass_ProgramID, "oldViewProj");
	GLuint numSamples_3p = glGetUniformLocation(thirdPass_ProgramID, "numSamples");
	GLuint light_source_position_3p = glGetUniformLocation(thirdPass_ProgramID,"light_source_pos_ScreenSpace");
	GLuint flare_radius_3p = glGetUniformLocation(thirdPass_ProgramID,"flare_radius");
	GLuint dtime_3p = glGetUniformLocation(thirdPass_ProgramID,"dTime");
	

    //Enable depth test
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

	/*camera.offsetOrientation(0,-90.0f);
	camera.setPosition(glm::vec3(40, 0, 0));*/
	camera.setAspectRatio((float)project::init_breite/(float)project::init_hoehe);

	lightCamera_pos_x.setFieldOfView(90.0f);
	lightCamera_neg_x.setFieldOfView(90.0f);
	lightCamera_pos_y.setFieldOfView(90.0f);
	lightCamera_neg_y.setFieldOfView(90.0f);
	lightCamera_pos_z.setFieldOfView(90.0f);
	lightCamera_neg_z.setFieldOfView(90.0f);


	lightCamera_pos_x.setOrientation(glm::mat4(glm::mat3(0, 0, -1, 0, -1, 0, -1, 0, 0)));
	lightCamera_neg_x.setOrientation(glm::mat4(glm::mat3(0, 0, 1, 0, -1, 0, 1, 0, 0)));
	lightCamera_pos_y.setOrientation(glm::mat4(glm::mat3(1, 0, 0, 0, 0, 1, 0, -1, 0)));
	lightCamera_neg_y.setOrientation(glm::mat4(glm::mat3(1, 0, 0, 0, 0, -1, 0, 1, 0)));
	lightCamera_pos_z.setOrientation(glm::mat4(glm::mat3(1, 0, 0, 0, -1, 0, 0, 0, -1)));
	lightCamera_neg_z.setOrientation(glm::mat4(glm::mat3(-1, 0, 0, 0, -1, 0, 0, 0, 1)));

	glm::vec3 m_directions[6];
	m_directions[0] = glm::vec3(1.0, 0.0, 0.0);
	m_directions[1] = glm::vec3(-1.0, 0.0, 0.0);
	m_directions[2] = glm::vec3(0.0, 1.0, 0.0);
	m_directions[3] = glm::vec3(0.0, -1.0, 0.0);
	m_directions[4] = glm::vec3(0.0, 0.0, 1.0);
	m_directions[5] = glm::vec3(0.0, 0.0, -1.0);

	glm::vec3 m_directions_env[6];
	m_directions_env[0] = glm::vec3(-1.0, 0.0, 0.0);
	m_directions_env[1] = glm::vec3(1.0, 0.0, 0.0);
	m_directions_env[2] = glm::vec3(0.0, -1.0, 0.0);
	m_directions_env[3] = glm::vec3(0.0, 1.0, 0.0);
	m_directions_env[4] = glm::vec3(0.0, 0.0, -1.0);
	m_directions_env[5] = glm::vec3(0.0, 0.0, 1.0);

	glm::vec3 m_up[6];
	m_up[0] = glm::vec3(0.0, -1.0, 0.0);
	m_up[1] = glm::vec3(0.0, -1.0, 0.0);
	m_up[2] = glm::vec3(0.0, 0.0, 1.0);
	m_up[3] = glm::vec3(0.0, 0.0, -1.0);
	m_up[4] = glm::vec3(0.0, -1.0, 0.0);
	m_up[5] = glm::vec3(0.0, -1.0, 0.0);

	GLenum error = glGetError();
	//SETUP FBO 1st renderpass
	GLuint fbo_1;
	GLuint tDepthCubeMap;
	GLuint tColorCubeMap;
	setupFBO(&fbo_1,&tColorCubeMap,&tDepthCubeMap);

	//SETUP FBO 1_2st renderpass
	GLuint fbo_12;
	GLuint tDepthCubeMap_2;
	GLuint tColorCubeMap_2;
	setupFBO(&fbo_12,&tColorCubeMap_2,&tDepthCubeMap_2);

	//SETUP FBO 2nd Renderpass
	GLuint fbo_2;
	glGenFramebuffers(1, &fbo_2);
	glBindFramebuffer(GL_FRAMEBUFFER, fbo_2);

	GLuint tDepthMap;
	glGenTextures(1, &tDepthMap);
	glBindTexture(GL_TEXTURE_2D, tDepthMap);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT24, rPass2Width, rPass2Height, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,GL_TEXTURE_2D, tDepthMap, 0);

	GLuint tColorMap;
    glGenTextures(1, &tColorMap);
    glBindTexture(GL_TEXTURE_2D, tColorMap);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);  
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F,rPass2Width, rPass2Height, 0, GL_RGB, GL_FLOAT, 0);
 
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,GL_TEXTURE_2D, tColorMap, 0);

	if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
		std::cout << "Framebuffer not complete: " << glCheckFramebufferStatus(GL_FRAMEBUFFER)<<"\n";
	//	return false;	
	}
	glBindFramebuffer(GL_FRAMEBUFFER, 0); 



	/* init_resources */
	GLfloat fbo_vertices[] = {
		-1, -1,
		 1, -1,
		-1,  1,
		 1,  1,
	};
	
	glGenBuffers(1, &vbo_fbo_vertices);
	glBindBuffer(GL_ARRAY_BUFFER, vbo_fbo_vertices);
	glBufferData(GL_ARRAY_BUFFER, sizeof(fbo_vertices), fbo_vertices, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	error = glGetError();
	std::cout<<error<<"\n";

	glUseProgram(firstPass_1_ProgramID);
	glUniform2fv(near_far_plane_11p,1,&glm::vec2(cm_cams_near_plane, cm_cams_far_plane)[0]);
	glUseProgram(firstPass_2_ProgramID);
	glUniform2fv(near_far_plane_12p,1,&glm::vec2(cm_cams_near_plane, cm_cams_far_plane)[0]);
	glUseProgram(secondPass_ProgramID);
	glUniform2fv(near_far_plane_2p,1,&glm::vec2(cm_cams_near_plane, cm_cams_far_plane)[0]);
	glUniform1i(tex_size_2p,shadowTexSideLength);
	glUniform1i(kernelsize_2p,4);
	glUniform1f(mipmapLevel_2p,0.5f);
	glm::mat4 oldVP = mat4();
	
	camera.setPosition(glm::vec3(20.5,2.2,8));
	camera.setOrientation(20.5,256);

	//camera.setPosition(glm::vec3(50,0,0));
	//camera.offsetOrientation(0,-90.0f);
	
	Animation<float> cam_orientation_anim(5,12);
	cam_orientation_anim.addValuePair(20.5,0);
	cam_orientation_anim.addValuePair(-30, 0);

	double prevTime = glfwGetTime();

	if (doAnimations) {
		wire_showWireFrame=false;
		float endTimeBlock = setupBlockAnimations(prevTime+5);
		endTimeBlock=setupPlaneAnimations(endTimeBlock);
		setupCameraMovement(prevTime);
		wireTogTime=endTimeBlock;
	}

    while(running)
    {
		//currTime in Seconds
		double currTime = glfwGetTime(); 
		//FPS
		double fps=0.0;
		if(fps_showFPS && (currTime > (fps_last_update+fps_update_interval_in_sec) || currTime < fps_update_interval_in_sec))
		{
			fps = ((double)fps_framecounter)/(currTime - fps_last_update);
			char title_buffer [50];
			sprintf(title_buffer,"%s FPS: %f",project::title.c_str(),fps);
			glfwSetWindowTitle(project::window,&title_buffer[0]);
			fps_framecounter=0;
			fps_last_update=currTime;	
		}
		else if(!fps_showFPS)
		{
			glfwSetWindowTitle(project::window,project::title.c_str());
		}

		//float erg = a.getValues(currTime)[index];
		
		
		float dTime = (float) currTime - prevTime;
		controls(dTime);
		prevTime = currTime;

		std::vector<Renderable>::iterator it;

		glm::mat4 VP = camera.matrix();
		
		mat4 cm_array [6];
		mat4 cm_view [6];
		
		glBindFramebuffer(GL_FRAMEBUFFER,fbo_1);
		
		for(int i = 0; i < 6; i++)
		{
			cm_array[i] = glm::perspective(90.0f, 1.0f, 0.3f, 100.0f) * glm::lookAt(Szene::szene[0].getPosition(), Szene::szene[0].getPosition() + m_directions[i], m_up[i]);
		}
		glUseProgram(firstPass_1_ProgramID);
		glUniformMatrix4fv(cm_mat_11p,	6, false, reinterpret_cast<const GLfloat*>(cm_array));
		glUniform3f(light_source_position_11p, Szene::szene[0].getPosition().x, Szene::szene[0].getPosition().y, Szene::szene[0].getPosition().z);

		glFramebufferTexture(GL_FRAMEBUFFER,GL_DEPTH_ATTACHMENT,tDepthCubeMap,0);
		glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, tColorCubeMap, 0);
		glClearColor(1,1,0,0);
		glClear( GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );
		glViewport(0, 0, shadowTexSideLength, shadowTexSideLength);

		if (doAnimations) {
			
			if (currTime < 15)
			{
				interpol = cam_orientation_anim.getValues(currTime);

				//std::cout << interpol[0] << " " << interpol[1] << std::endl;
				updateCameraPosition(currTime, interpol[0], interpol[1]);
			}
			else {
				updateCameraPosition(currTime, 0, 0);
			}

			int c = animationVec.size();
			for (int i = 0; i < c&& c > 0; i++)
			{
				Animation<vec3> a = animationVec[i];

				if (a.isRunning()) {	
					vec3 pos = a.getValues(currTime)[0];
					a.getObject()->setPosition(pos);
					a.getObject()->setVisible(true);
				}

				if(a.isDone()) {
					vec3 pos = a.getValues(currTime)[0];
					a.getObject()->setPosition(pos);
					a.getObject()->setVisible(true);
					animationVec.erase(animationVec.begin()+i);
					c--;
				}
			}
			c=planeAnimationVec.size();
			for (int i = 0; i < c&& c > 0; i++)
			{
				
				Animation<vec3> a = planeAnimationVec[i];
				if (a.isRunning()) {	
					vec3 pos = a.getValues(currTime)[0];
					a.getObject()->setPosition(pos);
				}
				
				
			}
			for (int i = 0; i < c&& c > 0; i++)
			{
				Animation<vec3> a = planeAnimationVec[i];

				if(a.isDone())
				{
					planeAnimationVec.erase(planeAnimationVec.begin()+i);
					c--;
				}
			}
			if(wireTogTime<currTime && wireTogTime+2>currTime)
			{
				wire_showWireFrame=true;
			}
			
			if(shadowTogStartTime<currTime && shadowTogStartTime+2>currTime)
			{
				varShadow_mode = 2;
			} 

			if(shadowTogEndTime<currTime && shadowTogEndTime+2>currTime)
			{
				varShadow_mode = 1;
			} 
		}
		glUniformMatrix4fv(trans_to_ref_11p, 1, GL_FALSE, &(mat4(1))[0][0]);		
		//1st RENDERPASS Schatten:
		for (it=Szene::szene.begin(); it<Szene::szene.end(); it++)
		{

			Renderable& r= (*it);
	
			//if(r.isTessellated())
			//	continue;
			if(!r.getDepthTest())
				continue;
			if(r.getName().compare(std::string("LICHT"))==0)
				continue;
			r.bindBuffers(0);
 
			glm::mat4 MVP = camera.projection() * camera.view() * r.getModelMatrix() ;

			glUniform3fv(CameraPositionID_11p,1,&(camera.position())[0]);
			glUniformMatrix4fv(ModelMatrixID_11p, 1, GL_FALSE, &(r.getModelMatrix())[0][0]);				

			r.render();	
				
		}

		glBindFramebuffer(GL_FRAMEBUFFER, 0);


		glBindFramebuffer(GL_FRAMEBUFFER,fbo_12);
		float tmp = camera.position().length();
		for(int i = 0; i < 6; i++)
		{
			cm_view[i] = glm::lookAt(Szene::szene[1].getPosition() , Szene::szene[1].getPosition() + m_directions[i], m_up[i]);
			cm_array[i] = glm::perspective(90.0f, 1.0f, 0.10f, 300.0f) * cm_view[i];
			
		}
		glUseProgram(firstPass_2_ProgramID);
		glUniformMatrix4fv(cm_mat_12p,	6, false, reinterpret_cast<const GLfloat*>(cm_array));
		glUniformMatrix4fv(ViewMatrixID_12p,	6, false, reinterpret_cast<const GLfloat*>(cm_view));

		glUniform3f(light_source_position_12p, Szene::szene[0].getPosition().x, Szene::szene[0].getPosition().y, Szene::szene[0].getPosition().z);

		glFramebufferTexture(GL_FRAMEBUFFER,GL_DEPTH_ATTACHMENT,tDepthCubeMap_2,0);
		glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, tColorCubeMap_2, 0);
		glClearColor(0,0,0,0);
		glClear( GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );
		glViewport(0, 0, shadowTexSideLength, shadowTexSideLength);
		glActiveTexture(GL_TEXTURE9);
		glBindTexture(GL_TEXTURE_CUBE_MAP, tDepthCubeMap);
		{
			glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);
			glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_COMPARE_FUNC, GL_LESS);
		}
		
		//1st RENDERPASS Evnvironment:
		glUniform1f(elapsedTimeID_12p,(float)currTime);
		//glUniform3fv(CameraPositionID_12p,1,&Szene::szene[1].getPosition()[0]);
		{
			glDisable(GL_DEPTH_TEST);
			glDepthMask(false);
			glUniform1f(useTexture_12p, 1.0);
			for(int i =0; i<5; i++)
			{
				Renderable r = Szene::Skybox[i];
				r.bindBuffers(1);
				glActiveTexture(GL_TEXTURE10);
				r.bindTexture();
				if(envMappingExtreme)
				{
					glUniform1f(tessellation_12p,r.isTessellated());
					glUniform3fv(tessellationParam_outer_12p,1,r.getTessellationParamsOuterVector());
					glUniform2fv(tessellationParam_inner_12p,1,r.getTessellationParamsInnerVector());
				}
				glUniformMatrix4fv(ModelMatrixID_12p, 1, GL_FALSE, &(r.getModelMatrix())[0][0]);
				r.render(envMappingExtreme);
			}
			glEnable(GL_DEPTH_TEST);
			glDepthMask(true);
		}
		for (it=Szene::szene.begin(); it<Szene::szene.end(); it++)
		{

			Renderable& r= (*it);
			if(r.getName().compare(Szene::szene[1].getName())==0)
				continue;
			r.bindBuffers(1);
			glActiveTexture(GL_TEXTURE10);
			r.bindTexture();
			
			if(envMappingExtreme)
			{
				glUniform1f(tessellation_12p,r.isTessellated());
				glUniform3fv(tessellationParam_outer_12p,1,r.getTessellationParamsOuterVector());
				glUniform2fv(tessellationParam_inner_12p,1,r.getTessellationParamsInnerVector());
			}
            glUniform1f(useTexture_12p, r.isTextured());			
            glUniformMatrix4fv(ModelMatrixID_12p, 1, GL_FALSE, &(r.getModelMatrix())[0][0]);
			
            r.render(envMappingExtreme);
		}

		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		//ENDE 1st Renderpass


		if(wire_showWireFrame)
		{
			glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
		}



		//2nd RENDERPASS
		

		glViewport(0, 0, rPass2Width, rPass2Height);
		if(doThirdPass)
			glBindFramebuffer(GL_FRAMEBUFFER, fbo_2);
		glUseProgram(secondPass_ProgramID);
		glUniform3f(light_source_position_2p, Szene::szene[0].getPosition().x, Szene::szene[0].getPosition().y, Szene::szene[0].getPosition().z);
		glUniform3fv(CameraPositionID_2p,1,&camera.position()[0]);
		//Subroutine
		//glUniformSubroutinesuiv(GL_FRAGMENT_SHADER,1, varShadow_mode==1?&subvar_2p:&subOmni_2p);
		glUniform1i(varShadow_2p,varShadow_mode);
		glActiveTexture(GL_TEXTURE9);
		glBindTexture(GL_TEXTURE_CUBE_MAP, tDepthCubeMap);
		if(varShadow_mode==1)
		{
			//ACHTUNG! Disable wenn in fragShader umstellung von samplercubeshadow auf samplerCube
			glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_COMPARE_MODE, GL_NONE);			
		}
		else
		{
			glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);
			glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_COMPARE_FUNC, GL_LESS);
		}
		glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
		glActiveTexture(GL_TEXTURE11);
		glBindTexture(GL_TEXTURE_CUBE_MAP, tColorCubeMap);
		glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
		glActiveTexture(GL_TEXTURE12);
		glBindTexture(GL_TEXTURE_CUBE_MAP, tColorCubeMap_2);
		glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
		glClearColor(0,0,0,0);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
		glUniform1f(elapsedTimeID_2p,(float)currTime);
		{
			glDisable(GL_DEPTH_TEST);
			glDepthMask(false);
			for(int i =0; i<5; i++)
			{
				Renderable r = Szene::Skybox[i];
				
				glUniform1i(reflection_flag_2p,r.getReflection());

				r.bindBuffers(1);
				glActiveTexture(GL_TEXTURE10);
				r.bindTexture();
			
				glUniform1f(useTexture_2p, r.isTextured());

				glUniform1f(tessellation_2p,r.isTessellated());
				glUniform3fv(tessellationParam_outer_2p,1,r.getTessellationParamsOuterVector());
				glUniform2fv(tessellationParam_inner_2p,1,r.getTessellationParamsInnerVector());

				glm::mat4 MVP = camera.projection() * camera.view() * r.getModelMatrix() ;
				glUniform3fv(CameraPositionID_2p,1,&(camera.position())[0]);
				//glUniform3f(CameraPositionID, 1, GL_FALSE, (camera.position())[0]);
				glUniformMatrix4fv(ProjectionMatrixID_2p, 1, GL_FALSE, &(camera.projection())[0][0]);
				glUniformMatrix4fv(ViewMatrixID_2p, 1, GL_FALSE, &(camera.view())[0][0]);
				glUniformMatrix4fv(ModelMatrixID_2p, 1, GL_FALSE, &(r.getModelMatrix())[0][0]);
				glUniformMatrix4fv(MatrixID_2p, 1, GL_FALSE, &MVP[0][0]);
				
				r.render(1);
			}
			glEnable(GL_DEPTH_TEST);
			glDepthMask(true);
		}
		for (it=Szene::szene.begin(); it<Szene::szene.end(); it++){
            Renderable& r= (*it);

			glUniform1i(reflection_flag_2p,r.getReflection());

            r.bindBuffers(1);
			glActiveTexture(GL_TEXTURE10);
			r.bindTexture();
			
            glUniform1f(useTexture_2p, r.isTextured());

			glUniform1f(tessellation_2p,r.isTessellated());
			glUniform3fv(tessellationParam_outer_2p,1,r.getTessellationParamsOuterVector());
			glUniform2fv(tessellationParam_inner_2p,1,r.getTessellationParamsInnerVector());

			glm::mat4 MVP = camera.projection() * camera.view() * r.getModelMatrix() ;
			glUniform3fv(CameraPositionID_2p,1,&(camera.position())[0]);
			//glUniform3f(CameraPositionID, 1, GL_FALSE, (camera.position())[0]);
			glUniformMatrix4fv(ProjectionMatrixID_2p, 1, GL_FALSE, &(camera.projection())[0][0]);
			glUniformMatrix4fv(ViewMatrixID_2p, 1, GL_FALSE, &(camera.view())[0][0]);
            glUniformMatrix4fv(ModelMatrixID_2p, 1, GL_FALSE, &(r.getModelMatrix())[0][0]);
            glUniformMatrix4fv(MatrixID_2p, 1, GL_FALSE, &MVP[0][0]);
			glUniform1f(elapsedTimeID_2p,(float)currTime);
            r.render(1);
			
        }/**/
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		//ENDE 2nd Renderpass
		glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
		//3rd Renderpass
		if(doThirdPass)
		{
			glViewport(0, 0, project::init_breite, project::init_hoehe);
			glm::mat4 IVP = glm::inverse(VP);
			glUseProgram(thirdPass_ProgramID);
			glViewport(0, 0, project::init_breite, project::init_hoehe);
			glClearColor(0.0, 0.0, 0.0, 0.0);
			glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
			
			glActiveTexture(GL_TEXTURE8);
			glBindTexture(GL_TEXTURE_2D, tColorMap);
			glActiveTexture(GL_TEXTURE7);
			glBindTexture(GL_TEXTURE_2D, tDepthMap);

			//Matrizen
			glUniformMatrix4fv(VP_3p, 1, GL_FALSE, &(VP)[0][0]);
			glUniformMatrix4fv(invVP_3p, 1, GL_FALSE, &(IVP)[0][0]);
			glUniformMatrix4fv(oldVP_3p, 1, GL_FALSE, &(oldVP)[0][0]);
		
			//Sample anzahl fr motionblur:
			glUniform1i(numSamples_3p, 4);
			glUniform1f(dtime_3p,dTime);

			//screenspace light
			vec4 light = VP*glm::vec4(Szene::szene[0].getPosition(),1);
			light/=light.w;
			glUniform3f(light_source_position_3p, light.x, light.y, light.z);
			glUniform1f(flare_radius_3p,0.03);
			glEnableVertexAttribArray(0);
 
			glBindBuffer(GL_ARRAY_BUFFER, vbo_fbo_vertices);
			glVertexAttribPointer(
			0,  // attribute
			2,                  // number of elements per vertex, here (x,y)
			GL_FLOAT,           // the type of each element
			GL_FALSE,           // take our values as-is
			0,                  // no extra data between each position
			0                   // offset of first element
			);
			glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
			glDisableVertexAttribArray(0);

			oldVP=VP;

		}
		//ENDE 3rd RenderPass

        //printf("Debug 3");
        //system("PAUSE");
		glDisableVertexAttribArray(0);
		glDisableVertexAttribArray(1);
		glDisableVertexAttribArray(2);
		glDisableVertexAttribArray(3);

		// Swap buffers
		glfwSwapBuffers(project::window);
		glfwPollEvents 	() ;
		fps_framecounter++;

        // exit if ESC was pressed or window was closed
        running = !(glfwGetKey(project::window, GLFW_KEY_ESCAPE) || glfwWindowShouldClose(project::window));
    }
	
	glDeleteBuffers(1, &vbo_fbo_vertices);
    glfwTerminate();

    return 0;
}

void project::init()
{
    glfwInit();
	project::window = glfwCreateWindow( project::init_breite, project::init_hoehe, project::title.c_str(),NULL,NULL);
	
	if( !project::window )
    {
        glfwTerminate();
        printf("Could not initialize GLFW-Context");
        exit(-10);
    }
	glfwMakeContextCurrent(project::window);
    GLenum err=glewInit();
    if(err!=GLEW_OK)
    {
		std::cerr<<glewGetErrorString(err);
        printf("Could not initialize GLEW-Context.");
        exit(err);

    }
	GLint max_att;
    glGetIntegerv(GL_MAX_VERTEX_ATTRIBS,&max_att);
    printf("GL_MAX_VERTEX_ATTRIBS:%d\n",max_att);
}

void setupFBO(GLuint *fbo, GLuint *tex, GLuint *depth)
{
	
	glGenFramebuffers(1, fbo);
	glBindFramebuffer(GL_FRAMEBUFFER, *fbo);
	
	// depth cube map
	
	glGenTextures(1, depth);
	glBindTexture(GL_TEXTURE_CUBE_MAP, *depth);
	/*glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );

	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );*/

	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_REPEAT );
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);  
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
	GLfloat maxA;
	glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maxA);
	//printf("\nGL_MAX_TEXTURE_MAX_ANISOTROPY_EXT: %f \n",maxA);
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAX_ANISOTROPY_EXT, maxA);
	//ACHTUNG! Disable wenn in fragShader umstellung von samplercubeshadow auf samplerCube
	
	for (uint face = 0; face < 6; face++) {
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_DEPTH_COMPONENT24, shadowTexSideLength, shadowTexSideLength, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);
	}
	
    glGenTextures(1, tex);
    glBindTexture(GL_TEXTURE_CUBE_MAP, *tex);
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );
	
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);  
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAX_ANISOTROPY_EXT, maxA);
       
	for (uint face = 0; face < 6; face++) {
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_RGB32F,shadowTexSideLength, shadowTexSideLength, 0, GL_RGB, GL_FLOAT, 0);
	}
 


	GLenum error = glGetError();
	std::cout<<error<<"\n";
	
	/*glDrawBuffer(GL_NONE);
	glReadBuffer(GL_NONE);*/
	
	if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
		std::cout << "Framebuffer not complete: " << glCheckFramebufferStatus(GL_FRAMEBUFFER)<<"\n";
	//	return false;	
	}
	
	glBindFramebuffer(GL_FRAMEBUFFER, 0);  
}
void controls(float elapsedTimeInSeconds)
{
	/*
    if(project::last_frame_time==0)
        project::last_frame_time=glfwGetTime();
    double now = glfwGetTime();
    glm::vec3 richtung=glm::vec3(
        sin(project::blickrichtung)*cos(project::neigungswinkel)   //mal cos(neigungswinkel) weil xrichtung im verhaeltnis kuerzer wird
        ,sin(project::neigungswinkel)
        ,cos(project::blickrichtung)*cos(project::neigungswinkel)     //mal cos(neigungswinkel) siehe x
      //,0.0f //richtung
    );
    double diff = (now-project::last_frame_time);

    glm::vec3 richtung_ = glm::vec3(
		sin(project::blickrichtung - PI/2.0f),
		0,
		cos(project::blickrichtung - PI/2.0f)
	);
	*/
	double x,y;
	
    glfwGetCursorPos(project::window,&x,&y);
	glfwGetWindowSize(project::window,&width,&height);
//	glfwSetMousePos(width/2,height/2);

	
	//printf("%d,%d\n",x,y);
    if (glfwGetKey(project::window, 'W' ) == GLFW_PRESS){
		 camera.offsetPosition(elapsedTimeInSeconds * moveSpeed * camera.forward());
	}
	else if(glfwGetKey(project::window, 'S' ) == GLFW_PRESS)
	{
		camera.offsetPosition(elapsedTimeInSeconds * moveSpeed * -camera.forward());
	}
	if (glfwGetKey(project::window, 'J' ) == GLFW_PRESS){
		vec3 pos = Szene::szene[0].getPosition();
		pos -= camera.right()* 1.2f * elapsedTimeInSeconds * moveSpeed;
		updateLightSourcePosition(pos);
		
	}
	else if(glfwGetKey(project::window, 'L' ) == GLFW_PRESS)
	{
		vec3 pos = Szene::szene[0].getPosition();
        pos += camera.right()*1.2f * elapsedTimeInSeconds * moveSpeed;
		updateLightSourcePosition(pos);
	}
	if (glfwGetKey(project::window, 'I' ) == GLFW_PRESS){
		vec3 pos = Szene::szene[0].getPosition();
		pos += camera.forward()*1.2f * elapsedTimeInSeconds * project::bewegungs_geschwindigkeit;
		updateLightSourcePosition(pos);
	}
	else if(glfwGetKey(project::window, 'K' ) == GLFW_PRESS)
	{
		vec3 pos = Szene::szene[0].getPosition();
		pos -= camera.forward()*1.2f * elapsedTimeInSeconds * project::bewegungs_geschwindigkeit;
		updateLightSourcePosition(pos);
	}
	if (glfwGetKey(project::window, 'A' ) == GLFW_PRESS){
		// project::position += richtung_ * (float)diff * project::bewegungs_geschwindigkeit;
		camera.offsetPosition(elapsedTimeInSeconds * moveSpeed * -camera.right());
	}
	else if(glfwGetKey(project::window, 'D' ) == GLFW_PRESS)
	{
        // project::position -= richtung_ * (float)diff * project::bewegungs_geschwindigkeit;
		camera.offsetPosition(elapsedTimeInSeconds * moveSpeed * camera.right());
	}

	if (glfwGetKey(project::window, 'Y' ) == GLFW_PRESS){
		// project::position += richtung_ * (float)diff * project::bewegungs_geschwindigkeit;
		camera.offsetPosition(elapsedTimeInSeconds * moveSpeed * -glm::vec3(0,1,0));
	}
	else if(glfwGetKey(project::window, 'X' ) == GLFW_PRESS)
	{
        // project::position -= richtung_ * (float)diff * project::bewegungs_geschwindigkeit;
		camera.offsetPosition(elapsedTimeInSeconds * moveSpeed * glm::vec3(0,1,0));
	}

	if (glfwGetKey(project::window, 'C' ) == GLFW_PRESS)
	{
		camera.offsetOrientation(elapsedTimeInSeconds * moveSpeed * 5, 0);
	}
	else if(glfwGetKey(project::window, 'V' ) == GLFW_PRESS)
	{
		camera.offsetOrientation(-moveSpeed * elapsedTimeInSeconds * 5, 0);
	}
	else if(glfwGetKey(project::window, 'F' ) == GLFW_PRESS)
	{
		fps_key_pressed=true;
	}
	else if(fps_key_pressed && glfwGetKey(project::window, 'F' ) == GLFW_RELEASE)
	{
		fps_key_pressed=false;
		fps_showFPS=!fps_showFPS;
	}
	else if(glfwGetKey(project::window, 'Q' ) == GLFW_PRESS)
	{
		wire_key_pressed=true;
	}
	else if(wire_key_pressed && glfwGetKey(project::window, 'Q' ) == GLFW_RELEASE)
	{
		wire_key_pressed=false;
		wire_showWireFrame=!wire_showWireFrame;
	}
	else if(glfwGetKey(project::window, 'O' ) == GLFW_PRESS)
	{
		varShadow_key_pressed=true;
	}
	else if(varShadow_key_pressed && glfwGetKey(project::window, 'O' ) == GLFW_RELEASE)
	{
		varShadow_key_pressed=false;
		varShadow_mode++;
		varShadow_mode %= 2;
	}
	
	if(glfwGetKey(project::window, GLFW_KEY_LEFT ) == GLFW_PRESS)
	{
		camera.offsetOrientation(0,elapsedTimeInSeconds * moveSpeed * -5 );
	}
	else if(glfwGetKey(project::window, GLFW_KEY_RIGHT ) == GLFW_PRESS)
	{
        camera.offsetOrientation(0,elapsedTimeInSeconds * moveSpeed * 5 );
	}

	if(glfwGetKey(project::window, 'H' ) == GLFW_PRESS)
	{
		std::cout<<camera.position().x<<","<<camera.position().y<<","<<camera.position().z<<endl;
		camera.printAngles();
	}
	/*else if(glfwGetKey(project::window, 'Z' ) == GLFW_PRESS)
	{
		animation_key_pressed=true;
	}
	else if(animation_key_pressed && glfwGetKey(project::window, 'Z' ) == GLFW_RELEASE)
	{
		animation_key_pressed=false;
		doAnimations=!doAnimations;
		if(doAnimations)
		{
			wire_showWireFrame=false;
			float endTimeBlock = setupBlockAnimations(glfwGetTime());
			endTimeBlock=setupPlaneAnimations(endTimeBlock);
			setupCameraMovement(glfwGetTime());
			wireTogTime=endTimeBlock+5;
		}
	}*/
   /* if(glfwGetKey(project::window, GLFW_KEY_UP ) == GLFW_PRESS)
	{
        project::neigungswinkel+=((float)diff)*project::bewegungs_geschwindigkeit*(PI/180);
	}
	else if(glfwGetKey(project::window, GLFW_KEY_DOWN ) == GLFW_PRESS)
	{
        project::neigungswinkel-=((float)diff)*project::bewegungs_geschwindigkeit*(PI/180);
	}
	*/
//    project::neigungswinkel += glm::sign(y/height)*project::bewegungs_geschwindigkeit*(PI/180);
 //   project::blickrichtung += glm::sign(x/width)*diff*project::bewegungs_geschwindigkeit*(PI/180);
	/*if(glfwGetKey(project::window,GLFW_KEY_ENTER) == GLFW_PRESS)
	{
	    printf("blickrichtung: %f,\nneigung: %f, Position: (%f,%f,%f)",project::blickrichtung,project::neigungswinkel,project::position.x,project::position.y,project::position.z);

	}
	project::ViewMatrix = glm::lookAt(
								project::position,           // Camera is here
								project::position+richtung, // and looks there : at the same position, plus "direction"
								glm::vec3(0.0f,1.0f,0.0f)   // Head is up (set to 0,-1,0 to look upside-down)
						   );
     // Projection matrix : 45 Field of View, 4:3 ratio, display range : 0.1 unit <-> 100 units
    float FoV = 45.0f;
	project::ProjectionMatrix = glm::perspective(FoV, 4.0f / 3.0f, 0.1f, 100.0f);

	project::last_frame_time = now;*/

}

void updateLightSourcePosition(glm::vec3 pos) {
		Szene::szene[0].setPosition(pos);
		lightCamera_pos_x.setPosition(pos);
		lightCamera_neg_x.setPosition(pos);
		lightCamera_pos_y.setPosition(pos);
		lightCamera_neg_y.setPosition(pos);
		lightCamera_pos_z.setPosition(pos);
		lightCamera_neg_z.setPosition(pos);/**/
}

float setupBlockAnimations(float startTime)
{
	float interval = 3.0f;
	float endTime = startTime + interval;
	
	Renderable* t_block = Szene::findRenderableByName("T");
	t_block->setVisible(false);
	vec3 t_block_pos = t_block->getPosition();
	Animation<vec3> t_anim(startTime, endTime, t_block);
	t_anim.addValuePair(vec3(t_block_pos.x,t_block_pos.y+20, t_block_pos.z), t_block_pos);
	
	startTime += interval;
	endTime += interval;
	
	Renderable* j_block = Szene::findRenderableByName("J");
	vec3 j_block_pos = j_block->getPosition();
	j_block->setVisible(false);
	Animation<vec3> j_anim(startTime, endTime, j_block);
	j_anim.addValuePair(vec3(j_block_pos.x, j_block_pos.y + 20, j_block_pos.z), j_block_pos);
	
	startTime += interval;
	endTime += interval;

	Renderable* z_block = Szene::findRenderableByName("Z");
	vec3 z_block_pos = z_block->getPosition();
	z_block->setVisible(false);
	Animation<vec3> z_anim(startTime, endTime, z_block);
	z_anim.addValuePair(vec3(z_block_pos.x, z_block_pos.y +20, z_block_pos.z), z_block_pos);
	
	startTime += interval;
	endTime += interval;

	Renderable* o_block = Szene::findRenderableByName("O");
	vec3 o_block_pos = o_block->getPosition();
	o_block->setVisible(false);
	Animation<vec3> o_anim(startTime, endTime, o_block);
	o_anim.addValuePair(vec3(o_block_pos.x, o_block_pos.y + 20, o_block_pos.z), o_block_pos);
	
	startTime += interval;
	endTime += interval;

	Renderable* t2_block = Szene::findRenderableByName("T2");
	vec3 t2_block_pos = t2_block->getPosition();
	t2_block->setVisible(false);
	Animation<vec3> t2_anim(startTime, endTime, t2_block);
	t2_anim.addValuePair(vec3(t2_block_pos.x,t2_block_pos.y+20, t2_block_pos.z), t2_block_pos);
	
	startTime += interval;
	endTime += interval;

	Renderable* j2_block = Szene::findRenderableByName("J2");
	vec3 j2_block_pos = j2_block->getPosition();
	j2_block->setVisible(false);
	Animation<vec3> j2_anim(startTime, endTime, j2_block);
	j2_anim.addValuePair(vec3(j2_block_pos.x, j2_block_pos.y + 20, j2_block_pos.z), j2_block_pos);
	
	startTime += interval;
	endTime += interval;

	Renderable* t3_block = Szene::findRenderableByName("T3");
	vec3 t3_block_pos = t3_block->getPosition();
	t3_block->setVisible(false);
	Animation<vec3> t3_anim(startTime, endTime, t3_block);
	t3_anim.addValuePair(vec3(t3_block_pos.x,t3_block_pos.y+20, t3_block_pos.z), t3_block_pos);
	
	startTime += interval;
	endTime += interval;

	Renderable* z2_block = Szene::findRenderableByName("Z2");
	vec3 z2_block_pos = z2_block->getPosition();
	z2_block->setVisible(false);
	Animation<vec3> z2_anim(startTime, endTime, z2_block);
	z2_anim.addValuePair(vec3(z2_block_pos.x, z2_block_pos.y +20, z2_block_pos.z), z2_block_pos);
	
	startTime += interval;
	endTime += interval;

	Renderable* j3_block = Szene::findRenderableByName("J3");
	vec3 j3_block_pos = j3_block->getPosition();
	j3_block->setVisible(false);
	Animation<vec3> j3_anim(startTime, endTime, j3_block);
	j3_anim.addValuePair(vec3(j3_block_pos.x, j3_block_pos.y + 20, j3_block_pos.z), j3_block_pos);
	
	endTime += 5;
	Renderable* water = Szene::findRenderableByName("Water");
	vec3 water_pos = water->getPosition();
	Animation<vec3> water_anim(startTime, endTime, water);
	water_anim.addValuePair(water_pos,  water_pos);
	

	animationVec.push_back(t_anim);
	animationVec.push_back(j_anim);
	animationVec.push_back(z_anim);
	animationVec.push_back(o_anim);
	animationVec.push_back(t2_anim);
	animationVec.push_back(j2_anim);
	animationVec.push_back(t3_anim);
	animationVec.push_back(z2_anim);
	animationVec.push_back(j3_anim);
	animationVec.push_back(water_anim);
	return endTime;
	
}

float setupPlaneAnimations(float startTime) {

	float interval = 3.0f;
	float endTime = startTime + interval;

	//Blick initial -x

	//2. Quadrant
	Renderable* plane = Szene::findRenderableByName("Plane");
	vec3 plane_pos = plane->getPosition();
	Animation<vec3> plane_anim(startTime, endTime, plane);
	plane_anim.addValuePair(plane_pos, plane_pos+vec3(-20,0,20));
	//,vec3(plane_pos.x-20, plane_pos.y, plane_pos.z-20));
	
	//1. Quadrant
	Renderable* plane2 = Szene::findRenderableByName("Plane2");
	vec3 plane2_pos = plane2->getPosition();
	Animation<vec3> plane2_anim(startTime, endTime, plane2);
	plane2_anim.addValuePair(plane2_pos, plane2_pos+vec3(-20,0,-20));
	
	//3. Quadrant
	Renderable* plane3 = Szene::findRenderableByName("Plane3");
	vec3 plane3_pos = plane3->getPosition();
	Animation<vec3> plane3_anim(startTime, endTime, plane3);
	plane3_anim.addValuePair(plane3_pos, plane3_pos+vec3(20,0,20));
	
	//4. Quadrant
	Renderable* plane4 = Szene::findRenderableByName("Plane4");
	vec3 plane4_pos = plane4->getPosition();
	Animation<vec3> plane4_anim(startTime, endTime, plane4);
	plane4_anim.addValuePair(plane4_pos,  plane4_pos+vec3(20,0,-20));

	float mag = 450;
	// 2. Quadrant
	Renderable* barrel = Szene::findRenderableByName("Barrel");
	vec3 barrel_pos = barrel->getPosition();
	Animation<vec3> barrel_anim(startTime, endTime, barrel);
	barrel_anim.addValuePair(barrel_pos,  barrel_pos+vec3(-mag,0,mag));

	// 3. Quadrant
	Renderable* barrel2 = Szene::findRenderableByName("Barrel2");
	vec3 barrel2_pos = barrel2->getPosition();
	Animation<vec3> barrel2_anim(startTime, endTime, barrel2);
	barrel2_anim.addValuePair(barrel2_pos,  barrel2_pos+vec3(mag,0,mag));

	// 1. Quadrant
	Renderable* barrel3 = Szene::findRenderableByName("Barrel3");
	vec3 barrel3_pos = barrel3->getPosition();
	Animation<vec3> barrel3_anim(startTime, endTime, barrel3);
	barrel3_anim.addValuePair(barrel3_pos,  barrel3_pos+vec3(-mag,0,-mag));

	// 4. Quadrant
	Renderable* barrel4 = Szene::findRenderableByName("Barrel4");
	vec3 barrel4_pos = barrel4->getPosition();
	Animation<vec3> barrel4_anim(startTime, endTime, barrel4);
	barrel4_anim.addValuePair(barrel4_pos,  barrel4_pos+vec3(mag,0,-mag));
	
	startTime+=5;
	endTime = startTime+10;
	Renderable* lamp = Szene::findRenderableByName("Streetlamp");
	vec3 lamp_pos = lamp->getPosition();
	Animation<vec3> lamp_anim(startTime, endTime, lamp);
	lamp_anim.addValuePair(lamp_pos, lamp_pos+vec3(0,200,0));
	
	planeAnimationVec.push_back(plane_anim);
	planeAnimationVec.push_back(plane2_anim);
	planeAnimationVec.push_back(plane3_anim);
	planeAnimationVec.push_back(plane4_anim);

	planeAnimationVec.push_back(barrel_anim);
	planeAnimationVec.push_back(barrel2_anim);
	planeAnimationVec.push_back(barrel3_anim);
	planeAnimationVec.push_back(barrel4_anim);

	planeAnimationVec.push_back(lamp_anim);
	


	return endTime;

	/*
	animationVec.push_back(plane_anim);
	animationVec.push_back(plane2_anim);
	animationVec.push_back(plane3_anim);
	animationVec.push_back(plane4_anim);*/

}

void setupCameraMovement(float startTime)
{

	shadowTogStartTime = startTime + 2;
	shadowTogEndTime = startTime + 4;

	startTime = shadowTogEndTime;

	Animation<vec3> cam_anim(startTime+1,+startTime+5);
	cam_anim.addValuePair(camera.position(),vec3(50,0,0));
	cam_anim.useCubicInterpolation();

	Animation<vec3> cam_anim2(startTime+5,+startTime+25);
	cam_anim2.addValuePair(vec3(50,0,0),vec3(30,40,30));
	cam_anim2.useCubicInterpolation();

	Animation<vec3> cam_anim3(startTime+25,startTime+29);
	cam_anim3.addValuePair(vec3(30,40,30),vec3(30,10,10));
	cam_anim3.useCubicInterpolation();

	Animation<vec3> cam_anim4(startTime+29,startTime+31);
	cam_anim4.addValuePair(vec3(30,10,10),vec3(40,10,-50));
	cam_anim4.useCubicInterpolation();

	Animation<vec3> cam_anim5(startTime+50,startTime+60);
	cam_anim5.addValuePair(vec3(40,10,-50),vec3(40,36,-50));
	cam_anim5.useCubicInterpolation();

	cameraMovementVec.push_back(cam_anim);
	cameraMovementVec.push_back(cam_anim2);
	cameraMovementVec.push_back(cam_anim3);
	cameraMovementVec.push_back(cam_anim4);
	cameraMovementVec.push_back(cam_anim5);
}

void updateCameraPosition(float currentTime, float up, float dir) 
{
	

	int c = cameraMovementVec.size();	
	for (int i = 0; i < c && c > 0; i++) 
	{
		Animation<vec3> a = cameraMovementVec[i];

		if(a.isRunning())
		{
			vec3 campos = a.getValues(currentTime)[0];
			
			//camera.offsetOrientation(up, dir);
			camera.offsetOrientation(up, dir);
			camera.setPosition(campos);

			//if (currentTime > 10)
				camera.lookAt(vec3(0));
			//glm::lookAt(camera.position(), vec3(0), vec3(up));
			
		}

		if(a.isDone()) {
			vec3 campos = a.getValues(currentTime)[0];
			camera.offsetOrientation(up, dir);
			camera.setPosition(campos);
			camera.lookAt(vec3(0));
			cameraMovementVec.erase(cameraMovementVec.begin()+i);
			c--;
		}
	}

}

