#include <stdio.h>
#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
using namespace std;


#include <stdlib.h>
#include <string.h>

#include <GL/glew.h>

#include "shader.hpp"

bool isTessellationAvailable()
{
	char* version = (char*) glGetString(GL_VERSION);
	if(version[0] >= '4')
		return true;
	return false;
	//std::cout<<version;
}
GLuint LoadShaders(const char * vertex_file_path, const char * tessellation_control_file_path, const char * tessellation_eval_file_path, const char * geometry_file_path,const char * fragment_file_path){
	//printf("test");
	// Create the shaders
	GLuint VertexShaderID = glCreateShader(GL_VERTEX_SHADER);
	GLuint TessControlShaderID;
	GLuint TessEvalShaderID;
	if(TESSELLATION)
	{	
		TessControlShaderID = glCreateShader(GL_TESS_CONTROL_SHADER);
		TessEvalShaderID = glCreateShader(GL_TESS_EVALUATION_SHADER);
	}
	GLuint GeometryShaderID;
	if(GEOMETRY)
		GeometryShaderID = glCreateShader(GL_GEOMETRY_SHADER);
	GLuint FragmentShaderID = glCreateShader(GL_FRAGMENT_SHADER);

	
	GLint Result = GL_FALSE;
	int InfoLogLength;


	GLuint ProgramID = glCreateProgram();

	// Compile Vertex Shader
	if(vertex_file_path!=0)
	{
		// Read the Vertex Shader code from the file
		std::string VertexShaderCode;
		std::ifstream VertexShaderStream(vertex_file_path, std::ios::in);
		if(VertexShaderStream.is_open()){
			std::string Line = "";
			while(getline(VertexShaderStream, Line))
				VertexShaderCode += "\n" + Line;
			VertexShaderStream.close();
		}

		printf("Compiling shader : %s\n", vertex_file_path);
		char const * VertexSourcePointer = VertexShaderCode.c_str();
		glShaderSource(VertexShaderID, 1, &VertexSourcePointer , NULL);
		glCompileShader(VertexShaderID);

		// Check Vertex Shader
		glGetShaderiv(VertexShaderID, GL_COMPILE_STATUS, &Result);
		glGetShaderiv(VertexShaderID, GL_INFO_LOG_LENGTH, &InfoLogLength);
		
		if(InfoLogLength > 0)
		{
			std::vector<char> VertexShaderErrorMessage(InfoLogLength);
			glGetShaderInfoLog(VertexShaderID, InfoLogLength, NULL, &VertexShaderErrorMessage[0]);
			fprintf(stdout, "%s\n", &VertexShaderErrorMessage[0]);
		}
		glAttachShader(ProgramID, VertexShaderID);
	}

	if(tessellation_control_file_path!=0 && tessellation_eval_file_path!=0 && TESSELLATION)
	{
		std::string TessControlShaderCode;
		std::string TessEvalShaderCode;
		//Read the TessControl Shader code from the file
		
		std::ifstream TessControlShaderStream(tessellation_control_file_path, std::ios::in);
		if(TessControlShaderStream.is_open()){
			std::string Line = "";
			while(getline(TessControlShaderStream, Line))
				TessControlShaderCode += "\n" + Line;
			TessControlShaderStream.close();
		}/**/

		//Read the TessEval Shader code from the file
		
		std::ifstream TessEvalShaderStream(tessellation_eval_file_path, std::ios::in);
		if(TessEvalShaderStream.is_open()){
			std::string Line = "";
			while(getline(TessEvalShaderStream, Line))
				TessEvalShaderCode += "\n" + Line;
			TessEvalShaderStream.close();
		}/**/

		// Compile TessControl Shader
		printf("Compiling shader : %s\n", tessellation_control_file_path);
		char const * TessControlSourcePointer = TessControlShaderCode.c_str();
		glShaderSource(TessControlShaderID, 1, &TessControlSourcePointer , NULL);
		glCompileShader(TessControlShaderID);

		// Check TessControl Shader
		glGetShaderiv(TessControlShaderID, GL_COMPILE_STATUS, &Result);
		glGetShaderiv(TessControlShaderID, GL_INFO_LOG_LENGTH, &InfoLogLength);
		
		if(InfoLogLength > 0)
		{
			std::vector<char> TessControlShaderErrorMessage(InfoLogLength);
			glGetShaderInfoLog(TessControlShaderID, InfoLogLength, NULL, &TessControlShaderErrorMessage[0]);
			fprintf(stdout, "%s\n", &TessControlShaderErrorMessage[0]);/**/
		}
		// Compile TessEval Shader
		printf("Compiling shader : %s\n", tessellation_eval_file_path);
		char const * TessEvalSourcePointer = TessEvalShaderCode.c_str();
		glShaderSource(TessEvalShaderID, 1, &TessEvalSourcePointer , NULL);
		glCompileShader(TessEvalShaderID);

		// Check TessEval Shader
		glGetShaderiv(TessEvalShaderID, GL_COMPILE_STATUS, &Result);
		glGetShaderiv(TessEvalShaderID, GL_INFO_LOG_LENGTH, &InfoLogLength);
		
		if(InfoLogLength > 0)
		{
			std::vector<char> TessEvalShaderErrorMessage(InfoLogLength);
			glGetShaderInfoLog(TessEvalShaderID, InfoLogLength, NULL, &TessEvalShaderErrorMessage[0]);
			fprintf(stdout, "%s\n", &TessEvalShaderErrorMessage[0]);/**/
		}
		glAttachShader(ProgramID, TessControlShaderID);
		glAttachShader(ProgramID, TessEvalShaderID);
	}

	if(geometry_file_path!=0 && GEOMETRY)
	{
		// Read the Geometry Shader code from the file
		std::string GeometryShaderCode;
		if(GEOMETRY)
		{
			std::ifstream GeometryShaderStream(geometry_file_path, std::ios::in);
			if(GeometryShaderStream.is_open()){
				std::string Line = "";
				while(getline(GeometryShaderStream, Line))
					GeometryShaderCode += "\n" + Line;
				GeometryShaderStream.close();
			} /**/
		}

		// Compile Geometry Shader
		printf("Compiling shader : %s\n", geometry_file_path);
		char const * GeometrySourcePointer = GeometryShaderCode.c_str();
		glShaderSource(GeometryShaderID, 1, &GeometrySourcePointer , NULL);
		glCompileShader(GeometryShaderID); /**/

		// Check Geometry Shader
		glGetShaderiv(GeometryShaderID, GL_COMPILE_STATUS, &Result);
		glGetShaderiv(GeometryShaderID, GL_INFO_LOG_LENGTH, &InfoLogLength);
		
		if(InfoLogLength > 0)
		{
			std::vector<char> GeometryShaderErrorMessage(InfoLogLength);
			glGetShaderInfoLog(GeometryShaderID, InfoLogLength, NULL, &GeometryShaderErrorMessage[0]);
			fprintf(stdout, "%s\n", &GeometryShaderErrorMessage[0]); /**/
		}
		glAttachShader(ProgramID, GeometryShaderID);
	}

	if(fragment_file_path!=0)
	{
		// Read the Fragment Shader code from the file
		std::string FragmentShaderCode;
		std::ifstream FragmentShaderStream(fragment_file_path, std::ios::in);
		if(FragmentShaderStream.is_open()){
			std::string Line = "";
			while(getline(FragmentShaderStream, Line))
				FragmentShaderCode += "\n" + Line;
			FragmentShaderStream.close();
		}

		// Compile Fragment Shader
		printf("Compiling shader : %s\n", fragment_file_path);
		char const * FragmentSourcePointer = FragmentShaderCode.c_str();
		glShaderSource(FragmentShaderID, 1, &FragmentSourcePointer , NULL);
		glCompileShader(FragmentShaderID);

		// Check Fragment Shader
		glGetShaderiv(FragmentShaderID, GL_COMPILE_STATUS, &Result);
		glGetShaderiv(FragmentShaderID, GL_INFO_LOG_LENGTH, &InfoLogLength);
		if(InfoLogLength > 0)
		{
			std::vector<char> FragmentShaderErrorMessage(InfoLogLength);
			glGetShaderInfoLog(FragmentShaderID, InfoLogLength, NULL, &FragmentShaderErrorMessage[0]);
			fprintf(stdout, "%s\n", &FragmentShaderErrorMessage[0]);
		}
		

		glAttachShader(ProgramID, FragmentShaderID);
	}
	
 
	// Link the program
	fprintf(stdout, "Linking program\n");
	glLinkProgram(ProgramID);

	// Check the program
	glGetProgramiv(ProgramID, GL_LINK_STATUS, &Result);
	glGetProgramiv(ProgramID, GL_INFO_LOG_LENGTH, &InfoLogLength);
	std::vector<char> ProgramErrorMessage( max(InfoLogLength, int(1)) );
	glGetProgramInfoLog(ProgramID, InfoLogLength, NULL, &ProgramErrorMessage[0]);
	fprintf(stdout, "%s\n", &ProgramErrorMessage[0]);


	glDeleteShader(VertexShaderID);
	if(TESSELLATION) {
		glDeleteShader(TessControlShaderID);
		glDeleteShader(TessEvalShaderID);
	}
	if(GEOMETRY)
		glDeleteShader(GeometryShaderID);
	glDeleteShader(FragmentShaderID);
	

	return ProgramID;
}


